﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.EvolutionTracking {
  [Item("EvolutionTrackingAnalyzer", "Base class for analyzers that use the genealogy graph")]
  [StorableClass]
  public abstract class EvolutionTrackingAnalyzer : SingleSuccessorOperator, IAnalyzer {
    #region parameter names
    private const string GenerationsParameterName = "Generations";
    private const string ResultsParameterName = "Results";
    private const string PopulationSizeParameterName = "PopulationSize";
    protected const string PopulationGraphResultName = "PopulationGraph";
    #endregion

    #region parameters
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public IValueParameter<IntValue> UpdateIntervalParameter {
      get { return (IValueParameter<IntValue>)Parameters["UpdateInterval"]; }
    }
    public ILookupParameter<IntValue> UpdateCounterParameter {
      get { return (ILookupParameter<IntValue>)Parameters["UpdateCounter"]; }
    }
    public ILookupParameter<IntValue> PopulationSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters[PopulationSizeParameterName]; }
    }
    #endregion

    #region properties
    public IntValue Generations { get { return GenerationsParameter.ActualValue; } }
    public ResultCollection Results { get { return ResultsParameter.ActualValue; } }
    public IntValue UpdateCounter { get { return UpdateCounterParameter.ActualValue; } }
    public IntValue UpdateInterval { get { return UpdateIntervalParameter.Value; } }
    public IntValue Generation { get { return GenerationsParameter.ActualValue; } }
    public IntValue PopulationSize { get { return PopulationSizeParameter.ActualValue; } }
    public IGenealogyGraph PopulationGraph {
      get {
        if (Results.ContainsKey(PopulationGraphResultName))
          return (IGenealogyGraph)Results[PopulationGraphResultName].Value;
        return null;
      }
    }

    #endregion

    public bool EnabledByDefault {
      get { return false; }
    }

    protected EvolutionTrackingAnalyzer() {
      Parameters.Add(new LookupParameter<IGenealogyGraph>(PopulationGraphResultName, "The genealogy graph."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "The number of generations."));
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection."));
      Parameters.Add(new ValueParameter<IntValue>("UpdateInterval", new IntValue(1)));
      Parameters.Add(new LookupParameter<IntValue>("UpdateCounter"));
      Parameters.Add(new LookupParameter<IntValue>(PopulationSizeParameterName, "The population size"));
    }

    protected EvolutionTrackingAnalyzer(EvolutionTrackingAnalyzer original, Cloner cloner)
      : base(original, cloner) {
    }

    [StorableConstructor]
    protected EvolutionTrackingAnalyzer(bool deserializing) : base(deserializing) { }
  }

  [Item("EvolutionTrackingAnalyzer", "Base class for analyzers that use the genealogy graph")]
  [StorableClass]
  public class EvolutionTrackingAnalyzer<T> : EvolutionTrackingAnalyzer where T : class, IItem {
    public EvolutionTrackingAnalyzer() { }

    protected EvolutionTrackingAnalyzer(EvolutionTrackingAnalyzer original, Cloner cloner) : base(original, cloner) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new EvolutionTrackingAnalyzer<T>(this, cloner);
    }

    [StorableConstructor]
    protected EvolutionTrackingAnalyzer(bool deserializing) : base(deserializing) { }

    public new IGenealogyGraph<T> PopulationGraph {
      get {
        if (Results.ContainsKey(PopulationGraphResultName))
          return (IGenealogyGraph<T>)Results[PopulationGraphResultName].Value;
        return null;
      }
    }
  }
}
